def importWave(filename):
    '''
    Return a tuple containing the framerate and audio signal samples from
    filename.  If filename is a stereo wave file, tuple is of length three.
    If filename is a mono wave file, tuple is of length two.

    importWave(filename) -> (samp_freq, signal1[, signal2])

    Input:
        filename - a string containing the name of a wave file to parse
    Output:
        samp_freq - the sampling frequency of the wave file
        signal1   - the sound from channel 1
        signal2   - if filename is stereo, the sound from channel 2
    Errors:
        'WaveIO Channel Error'        - if filename's channels is not 1 or 2
        'WaveIO Sample Width Error'   - if filename doesn't have 8, 16 or 32
                                        bit samples
        wave or struct module errors (i.e. if filename is compressed)
    '''

    import wave
    import struct

    #open wave file, get parameters and check for errors
    #if an error is raised, close the file and throw the error
    f = wave.open(filename)
    
    samp_width = f.getsampwidth()
    if not (samp_width in [1,2,4]):
        f.close()
        raise 'WaveIO Sample WidthError',\
              'Sample width of %s unparseable' % sampwidth
    samp_channels = f.getnchannels()
    if not (samp_channels in [1,2]):
        f.close()
        raise 'WaveIO Channel Error','%s channels unparseable' % channels
    samp_freq = f.getframerate()
    samp_length = f.getnframes()

    conv_code = packing_code( samp_width )

    signal1 = []
    signal2 = []

    frames = f.readframes(samp_length)
    f.close()
    for i in range(samp_length):         #get frame, unpack & append to signals
        frame = frames[ i   *samp_width*samp_channels:
                           (i+1)*samp_width*samp_channels]
        if samp_channels == 1:
            signal1.append( struct.unpack( conv_code, frame ))
        if samp_channels == 2:
            signal1.append( struct.unpack( conv_code, frame[0:samp_width] ))
            signal2.append( struct.unpack( conv_code,
                                           frame[samp_width:2*samp_width]))

    if samp_channels == 1:
        return (samp_freq, signal1)
    else:
        return (samp_freq, signal1, signal2)

def exportWave(filename, samp_rate, signal, signal2=[], samp_width=2, volume=1):
    '''
    Return True on successful write of audio data in signal to filename.  If
    signal2 is passed and the same length as signal, a stereo file is written.

    Inputs:
        filename   - name of wave file to be written
        samp_rate  - the sampling rate
        signal     - audio signal to be written
        signal2    - if signal2 is passed, length is checked against signal
                     and a stereo wave file is written
        samp_width - the sample size to be written, defaults to two bytes
        volume     - the volume scalar, should meet 0 < volume <= 1
    Outputs:
        True on success
    Raises:
        Possibly errors from the wave module.
        'WaveIO Signal Error'         - if len(signal) != len(signal2) w/
                                        passed signal2
        'WaveIO Sample Width Error'   - if samp_width not in [1,2,4]
        'WaveIO Sample Freq Error'    - if samp_rate <= 0 or not an int
    '''

    import wave, struct

    samp_channels = 1

    if not (signal2 == []):
        samp_channels = 2
        if len(signal) != len(signal):
            raise 'WaveIO Signal Error',\
                  'different sample lengths passed to waveio.exportWave'

    if not (samp_width in [1,2,4]):
        raise 'WaveIO Sample Width Error',\
              'Sample width of %s not valid - need 1, 2, or 4' % samp_width

    if (samp_rate <= 0) or (samp_rate != int(samp_rate)):
        raise 'WaveIO Sample Freq Error',\
              'Sample frequency of %s invalid' % samp_rate

    signal = scale_signal(signal, samp_width * 8, volume)
    if samp_channels == 2:
        signal2 = scale_signal(signal2, samp_width*8, volume)
    if samp_width != 1:
        # need to center non-8bit signals about zero
        signal =  map( lambda x: x - 2**(samp_width*8 - 1), signal )
        signal2 = map( lambda x: x - 2**(samp_width*8 - 1), signal2 )        

    f = wave.open(filename,'w')
    conv_code = packing_code( samp_width )
    f.setnchannels( samp_channels )
    f.setframerate( samp_rate )
    f.setsampwidth( samp_width )

    for i in range(len(signal)):
        if samp_channels   == 1:
            f.writeframes( struct.pack( conv_code, signal[i] ) )
        elif samp_channels == 2:
            f.writeframes( struct.pack( conv_code, signal[i] ) + \
                           struct.pack( conv_code, signal2[i]))
    return True

def packing_code(samp_width):
    """
    Return the code associated struct.[un]pack for the specificied number
    of bytes.
    Inputs:
        samp_width - number of bytes to [un]pack
    Outputs:
        the conversion code used with stuct.[un]pack
    Raises:
        'WaveIO Packing Error' - if the number of bytes not in 1, 2, 4
    """
    if samp_width == 1:               #8 bits are unsigned, 16 & 32 signed
        return 'B'     #unsiged 8 bits
    elif samp_width == 2:
        return 'h'     #signed 16 bits
    elif samp_width == 4:
        return 'i'     #signed 32 bits

    raise 'WaveIO Packing Error','Not able to parse %s bytes' % samp_width

def scale_signal(sig, bt, scalar):
    """
    Take an arbitrary signal, sig, stored as a list and return a signal
    that is integral and can be compressed in bt bits.  The signal is
    recentered about zero.
    Input:
        sig    - signal in a list
        bt     - bits to express signal in
        scalar - amount to scale signal by (if > 1, makes signal take more
                 than bt bits)
    Output:
        a list containing the linearly scaled, recentered signal ranging from
        0 to 2**(bt)
    Raises:
        rounding errors
        non-number cast errors
    """
    tmp = list( sig )
    min_tmp = min(tmp)
    max_tmp = max(tmp)
    scalar = (2**bt - 1.0) / (1.0 * (max_tmp - min_tmp)) * scalar
    tmp = map( lambda x: int(scalar*(x-min_tmp)), tmp)
    return tmp
