
import os
import sys
import math
import time

import struct


class option:
    def __init__(self):
        self.verbose=False
        self.debug=False

opt=option()        

def i2bs(x):
    '''Convert a (32-bit) int to a list of 4 byte values, e.g.
    i2bs(0xdeadbeef) = [222,173,190,239]
    12bs(0x4)        = [0,0,0,4]'''
    out=[]
    while x or len(out) < 4:
        out = [x & 0xff] + out
        x = x >> 8
    return out


buttonmap = {
    '2': 0x0001,
    '1': 0x0002,
    'B': 0x0004,
    'A': 0x0008,
    '-': 0x0010,
    'H': 0x0080,
    'L': 0x0100,
    'R': 0x0200,
    'D': 0x0400,
    'U': 0x0800,
    '+': 0x1000,
}

# BLUH. These should be less C-ish.
CMD_SET_REPORT = 0x52

RID_LEDS = 0x11
RID_MODE = 0x12
RID_IR_EN = 0x13 
RID_SPK_EN = 0x14
RID_STATUS = 0x15
RID_WMEM = 0x16
RID_RMEM = 0x17
RID_SPK = 0x18
RID_SPK_MUTE = 0x19
RID_IR_EN2 = 0x1a

MODE_BASIC = 0x30
MODE_ACC = 0x31
MODE_IR = 0x32
MODE_FULL = 0x3e
MODE_NUN= 0x37
#Once the Nunchuk is initialized, reports 0x32, 0x34, 0x35, 0x36, 0x37 and 0x3D will contain the six bytes of Nunchuk data (the actual offset depends on the report ID). 



IR_MODE_OFF =  0
IR_MODE_STD =  1
IR_MODE_EXP =  3
IR_MODE_FULL = 5

FEATURE_DISABLE = 0x00
FEATURE_ENABLE = 0x04

# Max value for IR dots
DOT_MAX = 0x3ff 

def decrypt(data):
    out=[]
    for i in range(len(data)):
        x=ord(data[i])
        out.append((x ^ 0x17) + 0x17)
    return out
        
def rotate(x,y,theta):
    '''Rotates the given (x,y) coordinates by theta radians around the center
    of the dots' view'''
    # Translate dot values so the center is (0,0)
    c=(DOT_MAX/2)
    x = c - x
    y = c - y
    # rotate about the center
    xprime = x*math.cos(theta) - y*math.sin(theta)
    yprime = x*math.sin(theta) + y*math.cos(theta)
    # now retranslate
    xprime = xprime + c
    yprime = yprime + c
    return (int(xprime),int(yprime))

class Wiimote(object):
    def __init__(self,fdin,fdout,number=0):
        self.done=False
        self.number=number
        self.mode       = 0
        self.ledmask    = 0
        self.buttonmask = 0
        
        self.force      = [0,0,0]
        self.force0 = [0,0,0]
        self.force_norm   = [1.,1.,1.]
        
        self.nforce      = [0,0,0]
        self.nforce0 = [0,0,0]
        self.nforce_norm   = [1.,1.,1.]
        
        self.njoy0= [0,0]
        self.njoy= [0,0]
        self.njoy_norm= [1.0,1.0]
        self.nbutton=0
        
        self.force_1g_diff = [0,0,0] # Difference between zero and 1g
        self.theta_g    = 0.0 # Angle of the remote with respect to gravity,
                            # calculated from the z-axis force. In radians.
        self.theta_g_x  = 0.0 # Same, but calculated from x-axis.

        self.dots       = [(DOT_MAX,DOT_MAX),(DOT_MAX,DOT_MAX)]
        self.theta      = 0.0 # dots' angle (again, in rad) from horizontal
        self.dotlist    = []  # a fifo queue of recent dots
        self.maxdots    = 10  # max length for dotlist<
        self.pointer = [0,0]  # Location of pointer. range is (0,DOT_MAX)

        self.rx = fdin
        self.cx = fdout
        self.nun=False


    def _handle_button_data(self,data):
        if len(data) != 4: return False
        # XXX: what's byte 1 for?
        newmask = (ord(data[2])<<8)+ord(data[3])
        # TODO: check newmask against current mask and send events?
        if newmask & buttonmap['H'] and not self.buttonmask & buttonmap['H']:
            print "Re-enabling IR"
            self.enable_IR()
        self.buttonmask = newmask
    
    def _handle_force_data(self,data):
        if len(data) != 3: return False
        force = [ord(d) for d in data]
        self.force = [nf*(f-f0) for f,f0,nf in zip(force,self.force0,self.force_norm)]
        return True

    def _handle_nunchuck_data(self,data):
        data=decrypt(data)
        #self.nforce = [ord(d) for d in data[2:-1]]
        force = [d for d in data[2:-1]]
        #print(force)
        self.nforce = [nf*(f-f0) for f,f0,nf in zip(force,self.nforce0,self.nforce_norm)]
        njoy= [d for d in data[:2]]
        self.njoy = [nf*(f-f0) for f,f0,nf in zip(njoy,self.njoy0,self.njoy_norm)]
        self.nbutton=data[-1] ^ 3
        #print(self.nforce)
        return True
    
    def _handle_IR_data(self,data):
        if len(data) != 6: return False
        if data ==' \xff'*6:
            self.dots=[(DOT_MAX,DOT_MAX),(DOT_MAX,DOT_MAX)]
        else:
            a,b,c,d,e,f = [ord(d) for d in data]
            # processing dots:
            # each tuple is 3 bytes in the form: x,y,extra
            # extra contains 8 bits of extra data as follows: [yyxxssss]
            # x and y are the high two bits for the full 10-bit x/y values.
            # s is some unknown info (size data?)
            x1=a+((c & 0x30) << 4) 
            y1=b+((c & 0xc0) << 2)
            x2=d+((f & 0x30) << 4)
            y2=e+((f & 0xc0) << 2)
            self.dots=[(x1,y1),(x2,y2)]
            self.dotlist.insert(0,self.dots)
            if len(self.dotlist) > self.maxdots:
                self.dotlist.pop()
        return True
    def _getpacket(self):
        data=self.rx.recv(1024)
        if len(data) == 4:    # button
            self._handle_button_data(data)
        elif len(data) == 7:  # button + accelerometer
            self._handle_button_data(data[0:4])
            self._handle_force_data(data[4:7])
        elif len(data) == 12:
            print("nunchuck??",data[7:12])
        elif len(data) == 23:
            #print("nunchuck??",data)
            self._handle_button_data(data[0:4])
            self._handle_force_data(data[4:7])
            self._handle_IR_data(data[7:13])
            self._handle_nunchuck_data(data[17:23])
        elif len(data) == 19: # button + accel + IR
            self._handle_button_data(data[0:4])
            self._handle_force_data(data[4:7])
            self._handle_IR_data(data[7:13])
            # I think the extra data is emitted if we see more than two dots
            extradata = data[13:19]
            if opt.debug and (extradata != "\xff"*len(extradata)):
                print "Interesting extradata: %s\n" % extradata.encode("hex")
        elif len(data) == 0:  # Wiimote went away!
            if opt.debug: print "Lost wiimote #%i" % (self.number+1)
            self.done = True
        else:
            print "Unknown packet len %i: 0x%s" % (len(data),data.encode("hex"))
    def setled(self,num):
        if opt.debug: print "setled(%i)" % num
        if num < 4:
            self.ledmask = self.ledmask | (0x10 << num)
            self._led_command()
    def clearled(self,num):
        if opt.debug: print "clearled(%i)" % num
        if num < 4:
            self.ledmask = self.ledmask & ~(0x10 << num)
            self._led_command()
    
    def showstatus(self):
        sys.stdout.write(self.status_str() + "\r")
        sys.stdout.flush()
    def set(self,mode):
        self.mode = mode
        # XXX wiimotulator.py has flags for setting 0x01 in the first byte for 
        # 'rmbl' and 0x04 for 'cont'. Both of these are always off.
        # No idea why.
        self._send_command(CMD_SET_REPORT,RID_MODE,[0,mode])
    def enable_force(self):
        self.set(self.mode | MODE_ACC)
        self.get_force_calibration()
    def enable_nun(self):
        self._write_mem(0x04A40040,[0x00])
        self.set(MODE_NUN)
        self.nun=True
        self.get_force_calibration()
        try:
            self.get_nforce_calibration()
        except:
            self.nun=False
            self.set(MODE_ACC)
    def enable_IR(self):
        self.set(self.mode | MODE_IR)
        self._send_command(CMD_SET_REPORT,RID_IR_EN,[FEATURE_ENABLE])
        self._send_command(CMD_SET_REPORT,RID_IR_EN2,[FEATURE_ENABLE])
        # Enable IR device
        self._write_mem(0x04b00030,[0x01])
        # Set sensitivity constants
        self._write_mem(0x04b00030,[0x08])
        self._write_mem(0x04b00006,[0x90])
        self._write_mem(0x04b00008,[0xc0])
        self._write_mem(0x04b0001a,[0x40])
        self._write_mem(0x04b00033,[0x33])
        # Enable IR data output
        self._write_mem(0x04b00030,[8])
        

    def _led_command(self):
        self._send_command(CMD_SET_REPORT,RID_LEDS,[self.ledmask])
    def _waitforpacket(self,header,max=32):
        r=''
        n=0
        while (n<max) and not r.startswith(header):
            r = self.rx.recv(1024)
            n = n + 1
        if opt.debug: print "Leaving _waitforpacket() after %i packets" % n
        if not r.startswith(header):
            return None
        else:
            return r
    def _waitforok(self):
        self._waitforpacket('\xa1\x22\x00')
    def _read_mem(self,offset,size):
        if size >= 16:
            print "ERROR: _read_mem can't handle size > 15 yet"
            return None
        # RMEM command wants: [offset,size]
        self._send_command(CMD_SET_REPORT,RID_RMEM,i2bs(offset)+[0,size])
        data = self._waitforpacket('\xa1\x21')
        if data:
            # TODO check error flag, continuation, etc
            return data[7:]
        else:
            return None
    def debug_nunchuck(self):
        addr=0x04a40020
        print("debugging")
        res=self._read_mem(addr,14)
        if res==None:
            print("Not good")
            return
        print("nf0 ",self.nforce0)
        print("nfn ",self.nforce_norm)
        
        print("f0 ",self.force0)
        print("fn ",self.force_norm)
        
            
    
    def get_force_calibration(self):
        data=[ord(b) for b in self._read_mem(0x16,10)]
        self.force0 = data[0:3]
        force_1g   = data[4:7]
        # XXX currently we don't know what data[3], data[7], or data[8:9] are
        for b in range(3):
            self.force_norm[b] = .2/(force_1g[b] - self.force0[b])
            
    
    def get_nforce_calibration(self):
        
        data=self._read_mem(0x04a40020,14)
        data=decrypt(data)
        #Falscher Messwert in z-Richtung
        print(data)
        #print(data)
        self.nforce0 = data[0:3]
        force_1g   = data[4:7]
        # XXX currently we don't know what data[3], data[7], or data[8:9] are
        for b in range(3):
            self.nforce_norm[b] = .2/(force_1g[b] - self.nforce0[b])
        normm=[]
        avv=[]
        temp =data[8:11]
        norm1=1./(temp[2]-temp[1])
        av=temp[2]
        normm.append(norm1)
        avv.append(av)
        temp =data[11:14]
        norm1=1./(temp[2]-temp[1])
        av=temp[2]
        normm.append(norm1)
        avv.append(av)
        self.njoy0=avv
        self.njoy_norm=normm
    
    def _write_mem(self,offset,data):
        # WMEM command wants: [offset,size,data]
        # offset = 32-bit, bigendian. data is 0-padded to 16 bytes.
        size = len(data)
        if size > 16: return False # Too much data!
        if size < 16: data = data + [0]*(16-size)
        self._send_command(CMD_SET_REPORT,RID_WMEM,i2bs(offset)+[size]+data)
        self._waitforok()
    def _send_command(self,cmd,report,data):
        if opt.debug: print "_send_command(%#x,%#x,%s)" % (cmd,report,data)
        self.cx.send(chr(cmd) + chr(report) + "".join([chr(d) for d in data]))
    def calc_theta_g(self):
        '''Use the z and x accelerometer values to figure out the wiimote's
        orientation with respect to gravity.'''
        # sanity - return if we have no calibration data
        if self.force_1g[0] == 0: return self.theta_g
        # rotating from face-up to upside-down, force[2] goes from 
        # force_1g[2] to force_zero[2]-force_1g[2]. The normal force of
        # gravity should be force_zero-force_1g - call this 'g'.
        # It seems intuitive that this should map to a cosine wave - we start
        # at 1g for face-up, then zero for a quarter-turn, -1g for half, etc.
        zg = float(self.force[2]-self.force_zero[2])/self.force_1g_diff[2]
        # If we're seeing more than 1g, probably this data isn't reliable
        # for determining orientation, so we ignore it
        if abs(zg) <= 1.0:
            self.theta_g = math.acos(zg)
        # Do the same thing with force[0] - it goes from 0->+/-1g->0, just like
        # a sine wave
        xg = float(self.force[0]-self.force_zero[0])/self.force_1g_diff[0]
        if abs(xg) <= 1.0:
            self.theta_g_x = math.asin(xg)
        # For convenience, return theta_g
        return self.theta_g

    def calc_pointer(self):
        '''Calculate the position of the pointer, taking into account the 
        rotation of the controller.
        Sets self.theta and self.pointer; returns self.pointer.'''
        # Credit for most of the math here goes to my esteemed colleague Mike
        # (mikem@redhat.com). Finally, all those years TA-ing Calc 1 are
        # paying off!
        # One of the dots is bogus/missing. Bail out.
        # TODO: keep track of the previous dot positions and guess instead of
        # immediately bailing out?
        if (DOT_MAX,DOT_MAX) in self.dots:
            return self.pointer
        ((x1,y1),(x2,y2)) = self.dots
        # FIXME: for some reason, py never goes above ~750.
        # Might be my bogus IR emitters (half-power every 15 degrees
        # away from center! Thanks, Radio Shack.) 
        # But it might also be that the IR camera is calibrated to 
        # assume the sensor bar should be on the bottom of the TV.
        # Since IR calibration is still Black Magick, I am forcing
        # a scale factor for y here.
        y1 = y1 * DOT_MAX / 760
        y2 = y2 * DOT_MAX / 760

        # Determine rotation angle. SOH CAH TOA ftw.
        if (x1 != x2):
            self.theta = math.atan(float(y2-y1)/float(x1-x2))
        else:
            self.theta = math.pi/2
            if y1 > y2:
                self.theta = -self.theta
        # If the accel. says we are upside-down, add half a turn to theta
        tg = math.degrees(self.calc_theta_g())
        if tg > 90.0:
            self.theta = self.theta+math.pi
        if tg < -90.0:
            self.theta = self.theta-math.pi
        # rotate dots around center by theta.
        (x1,y1) = rotate(x1,y1,self.theta)
        (x2,y2) = rotate(x2,y2,self.theta)
        # They should now be horizontal (y1 should be very close to y2).
        # Average the two X values (find the center between them)
        px = (x1+x2)/2
        # Horizontal means y1 = y2, so there's no need to average them.
        # In fact, let's output an error message if the rotate messed up.
        if y2 != y1:
            print "post-rotation Y delta=%i" % abs(y1-y2)
        # We do need to flip the incoming y data.
        py = DOT_MAX - y1

        # Do some scaling - ignore the outer edges of the screen
        # FIXME: fix scaling such that the center of the wiimote image
        # maps to the top of the screen
        # Center point of the screen is (c,c)
        c = DOT_MAX/2
        maxd = 0.33 * DOT_MAX # max allowable distance from center
        # If this point is less than (maxd) from the center of the image,
        # draw it.
        if (abs(px-c) <= maxd) and (abs(py-c) <= maxd):
            # px/py are in the range [c-maxd,c+maxd]
            px = px - (c-maxd)
            py = py - (c-maxd)
            # Now they're in the range [0,2*maxd]. Scale to DOT_MAX.
            px = px * (DOT_MAX/(2*maxd))
            py = py * (DOT_MAX/(2*maxd))
            # Hooray! We did it!
            self.pointer = [int(px),int(py)]
        return self.pointer
    def pointer_str(self):
        return "(%4i,%4i)" % (self.pointer[0],self.pointer[1])


